import dataclasses
import json
import logging
import os
import pickle
import typing as tp
from functools import lru_cache, partial
from logging.config import valid_ident
from os.path import join

import flax.linen as nn
import jax
import jax.numpy as jnp
import numpy as np
from flax.core.frozen_dict import freeze, unfreeze
from flax.training.train_state import TrainState

import lacss.modules
import lacss.train.strategy as strategy
from lacss.ops import patches_to_label
from lacss.train import Inputs

_cached_partial = lru_cache(partial)

Image = tp.Any

model_urls = {
    "livecell": "https://data.mendeley.com/public-files/datasets/sj3vrvm6w3/files/439e524f-e4e9-4f97-9f38-c22cb85adbd1/file_downloaded",
    "tissuenet": "https://data.mendeley.com/public-files/datasets/sj3vrvm6w3/files/1e0a839d-f564-4ee0-a4f3-34792df7c613/file_downloaded",
}


def load_from_pretrained(pretrained):
    if os.path.isdir(pretrained):
        import json

        try:
            with open(join(pretrained, "config.in")) as f:
                cfg = json.load(f)
        except:
            raise ValueError(
                f"Cannot open 'config.in' in {pretrained} as a json config file."
            )
        with open(join(pretrained, "weights.pkl")) as f:
            params = pickle.load(f)

    else:
        if os.path.isfile(pretrained):
            with open(pretrained, "rb") as f:
                thingy = pickle.load(f)

        else:
            from urllib.request import Request, urlopen
            import io

            headers = {"User-Agent": "Wget/1.13.4 (linux-gnu)"}
            req = Request(url=pretrained, headers=headers)

            bytes = urlopen(req).read()
            thingy = pickle.loads(bytes)

        if isinstance(thingy, lacss.train.Trainer):
            module = thingy.model
            params = thingy.params

            if not isinstance(module, lacss.modules.Lacss):
                module = module.bind(dict(params=params))
                module, params = module.lacss.unbind()
                params = params["params"]

            cfg = module

        else:
            cfg, params = thingy

    if "params" in params and len(params) == 1:
        params = params["params"]

    # making a round-trip of module->cfg->module to icnrease backward-compatibility
    # if isinstance(cfg, nn.Module):

    #     cfg = dataclasses.asdict(cfg)

    # module = lacss.modules.Lacss(**cfg)

    return cfg, freeze(params)


@partial(jax.jit, static_argnums=0)
def _predict(apply_fn, params, image):
    return apply_fn(dict(params=params), image)


class Predictor:
    """ Main interface for model deployment.

    Attributes:
        module: The underlying FLAX module 
        params: Model weights.
        detector: The detector submodule for convinence. It is common to 
            modidify this submodule to find optimal parameters for detection.
            Note this submodule is weight-less.
    
    Constructor:
        url: A URL of the saved model. URLs for build-in pretrained models can be found
            in lacss.deploy.model_urls
        precompile_shape: Optionally supply the image shape for precompiling. Otherwise 
            the model will be recompiled for every new input image shape.
    
    Usage Example:
        url = lacss.deploy.model_urls["livecell"]
        predictor = lacss.deploy.Predictor(url)
        label = predictor.predict_label(image)
    """
    def __init__(self, url, precompile_shape=None):
        self.model = load_from_pretrained(url)

        if precompile_shape is not None:
            logging.info("Prcompile the predictor for image shape {precompile_shape}.")
            logging.info("This will take several minutes.")

            try:
                iter(precompile_shape[0])
            except:
                precompile_shape = [precompile_shape]
            for shape in precompile_shape:
                x = np.zeros(shape)
                _ = self.predict(x)

    def __call__(self, inputs, **kwargs):
        """ Perform model prediction
        """
        inputs_obj = Inputs.from_value(inputs)

        return self.predict(*inputs_obj.args, **inputs_obj.kwargs, **kwargs)

    def predict(
        self,
        image: Image,
        min_area: float = 0,
        remove_out_of_bound: bool = False,
        scaling: float = 1.0,
        **kwargs,
    ):
        """Predict segmentation.

        Args:
            image: A ndarray of (h,w,c) format. Value of c must be 1-3
            min_area: Minimum area of a valid prediction. Default is 0
            remove_out_of_bound: Whether to remove out-of-bound predictions. Default is False.
            scaling: A image scaling factor. If not 1, the input image will be resized internally before fed
                to the model. The results will be resized back to the scale of the orginal input image.

        Returns:
            Raw model prediction as a dict. Most important keys are

            instance_output: Segmentation instances. (N, S, S) array
            instance_yc: The meshgrid y-coordinates of the instances. (N, S, S) array
            instance_xc: The meshgrid x-coordinates of the instances. (N, S, S) array
            pred_scores: The prediction scores of each instance. (N) array. -1 indicating invalid predictions.
            pred_locations: The prediction centroids of each instances. (N, 2) array
        """

        module, params = self.model

        if len(kwargs) > 0:
            apply_fn = _cached_partial(module.apply, **kwargs)
        else:
            apply_fn = module.apply

        if scaling != 1.0:
            h, w, c = image.shape
            image = jax.image.resize(
                image, [round(h * scaling), round(w * scaling), c], "linear"
            )

        preds = _predict(apply_fn, params, image)
        del preds["encoder_features"]
        del preds["decoder_features"]
        del preds["lpn_features"]
        del preds["lpn_scores"]
        del preds["lpn_regressions"]

        if min_area > 0:
            areas = jnp.count_nonzero(preds["instance_logit"] > 0, axis=(1, 2))
            preds["pred_scores"] = jnp.where(
                areas > min_area,
                preds["pred_scores"],
                -1,
            )

        if remove_out_of_bound:
            pred_locations = preds["pred_locations"]
            h, w, _ = image.shape
            valid_locs = (pred_locations >= 0).all(axis=-1)
            valid_locs &= pred_locations[:, 0] < h
            valid_locs &= pred_locations[:, 1] < w
            preds["pred_locations"] = jnp.where(
                valid_locs[:, None],
                preds["pred_locations"],
                -1,
            )
            preds["pred_scores"] = jnp.where(
                valid_locs,
                preds["pred_scores"],
                -1,
            )

        if scaling != 1.0:
            (
                preds["instance_output"],
                preds["instance_yc"],
                preds["instance_xc"],
            ) = lacss.ops.rescale_patches(preds, 1 / scaling)
            preds["pred_locations"] = preds["pred_locations"] / scaling

        return preds

    def predict_label(self, image, *, score_threshold=0.5, **kwargs):
        """ A convinent funciton to obtain prediction in image label format

        Args:
            image: Input image.
            score_threshold: The minimal prediction scores. Default is 0.5

        Returns:
            label: A [H, W] array. The values indicate the id of the cells. For cells
                with overlapping areas, the pixel is assigned for the cell with higher
                prediction scores.
        """
        preds = self.predict(image, **kwargs)

        return patches_to_label(
            preds, 
            input_size=image.shape[:2], 
            score_threshold=score_threshold,
        )

    @property
    def module(self):
        return self.model[0]
    
    @property
    def params(self):
        return self.model[1]
    
    @params.setter
    def params(self, new_params):
        self.model = self.module, new_params

    @property
    def detector(self):
        return self.module.detector

    # FIXME Change the detector without first compile the mode will result
    # in error. This is a hidden gotcha. Need to setup warning to the user.
    @detector.setter
    def detector(self, new_detector):
        from .modules import Detector

        cur_module = self.module

        if isinstance(new_detector, dict):
            new_detector = Detector(**new_detector)

        if isinstance(new_detector, Detector):
            self.model = (
                lacss.modules.Lacss(
                    self.module.backbone,
                    self.module.lpn,
                    new_detector,
                    self.module.segmentor,
                ),
                self.model[1],
            )
        else:
            raise ValueError(f"{new_detector} is not a Lacss Detector")



from .auxiliary import *
from .convnext import ConvNeXt
from .detector import Detector
from .lacss import Lacss
from .lpn import LPN
from .resnet import ResNet
from .segmentor import Segmentor
from .unet import UNet



from functools import partial

import jax
import jax.numpy as jnp
import optax

from ..ops import sorbel_edges
from ..train.loss import Loss
from .detection import _binary_focal_crossentropy

EPS = jnp.finfo("float32").eps


def _compute_edge(instance_output, instance_yc, instance_xc, height, width):
    _, ps, _ = instance_yc.shape
    padding = ps // 2 + 1

    patch_edges = jnp.square(sorbel_edges(instance_output))
    patch_edges = (patch_edges[0] + patch_edges[1]) / 8.0
    patch_edges = jnp.sqrt(jnp.clip(patch_edges, 1e-8, 1.0))  # avoid GPU error
    # patch_edges = jnp.where(patch_edges > 0, jnp.sqrt(patch_edges), 0)
    combined_edges = jnp.zeros([height + padding * 2, width + padding * 2])
    combined_edges = combined_edges.at[
        instance_yc + padding, instance_xc + padding
    ].add(patch_edges)
    combined_edges = combined_edges[
        padding : padding + height, padding : padding + width
    ]
    combined_edges = jnp.tanh(combined_edges)

    return combined_edges


def self_supervised_edge_loss(preds, inputs, **kwargs):
    """LACSS cell border consistency loss"""

    instance_output = preds["instance_output"]
    instance_yc = preds["instance_yc"]
    instance_xc = preds["instance_xc"]
    height, width, _ = inputs["image"].shape
    instance_edge = _compute_edge(
        instance_output, instance_yc, instance_xc, height, width
    )
    instance_edge_pred = jax.nn.sigmoid(preds["edge_pred"])

    return optax.l2_loss(instance_edge_pred, instance_edge).mean()


def self_supervised_segmentation_loss(
    preds,
    inputs,
    *,
    offset_sigma=jnp.array([10.0]),
    offset_scale=jnp.array([2.0]),
    **kwargs,
):
    """LACSS image segmentation consistenct loss"""

    height, width, _ = inputs["image"].shape
    _, ps, _ = preds["instance_yc"].shape

    def _max_merge(pred):
        label = jnp.zeros([height + ps, width + ps]) - 1.0e6
        yc, xc = pred["instance_yc"], pred["instance_xc"]
        logit = pred["instance_logit"]

        if "category" in inputs and inputs["category"] is not None:
            c = inputs["category"].astype(int).squeeze()
        else:
            c = 0

        sigma = offset_sigma[c]
        scale = offset_scale[c]
        offset = (jnp.mgrid[:ps, :ps] - (ps - 1) / 2) ** 2 / (2 * sigma * sigma)
        offset = jnp.exp(-offset.sum(axis=0)) * scale
        logit += offset

        label = label.at[yc + ps // 2, xc + ps // 2].max(logit)
        label = label[ps // 2 : -ps // 2, ps // 2 : -ps // 2]

        return label

    # fg_pred = jax.nn.tanh(preds["fg_pred"] / 2)
    # fg_label = jax.nn.tanh(_max_merge(preds) / 2)
    fg_pred = jax.nn.sigmoid(preds["fg_pred"])
    fg_label = jax.lax.stop_gradient(jax.nn.sigmoid(_max_merge(preds)))

    locs = jnp.floor(inputs["gt_locations"] + 0.5)
    locs = locs.astype(int)
    fg_label = fg_label.at[locs[:, 0], locs[:, 1]].set(1.0)

    assert fg_pred.shape == fg_label.shape

    # loss = 1.0 - fg_pred * jax.lax.stop_gradient(fg_label)
    loss = (1.0 - fg_label) * fg_pred + fg_label * (1.0 - fg_pred)

    loss = loss.mean()

    return loss


def aux_size_loss(preds, inputs, *, weight=0.01, **kwargs):
    """auxillary loss to prevent model collapse"""

    height, width, _ = inputs["image"].shape

    valid_locs = (
        (preds["instance_yc"] >= 0)
        & (preds["instance_yc"] < height)
        & (preds["instance_xc"] >= 0)
        & (preds["instance_xc"] < width)
    )

    areas = jnp.sum(
        preds["instance_output"], axis=(-1, -2), where=valid_locs, keepdims=True
    )
    areas = jnp.clip(areas, EPS, 1e6)
    loss = jax.lax.rsqrt(areas) * preds["instance_output"].shape[-1]

    mask = preds["instance_mask"]
    n_instances = jnp.count_nonzero(mask) + EPS
    loss = loss.sum(where=mask) / n_instances

    return loss * weight


def supervised_segmentation_loss(preds, labels, **kwargs):

    if "gt_labels" in labels:

        mask = (labels["gt_labels"] > 0).astype("float32")

    else:

        mask = labels["gt_mask"].astype("float32")

    return optax.sigmoid_binary_cross_entropy(preds["fg_pred"], mask).mean()


class AuxEdgeLoss(Loss):
    def call(
        self,
        inputs: dict,
        preds: dict,
        **kwargs,
    ):
        if not "training_locations" in preds:
            return 0.0

        return self_supervised_edge_loss(preds, inputs)


class AuxSizeLoss(Loss):
    def __init__(self, alpha=1e-2):
        super().__init__()
        self.a = alpha

    def call(self, inputs, preds, **kwargs):

        return aux_size_loss(preds, inputs, weight=self.a)


class AuxSegLoss(Loss):
    def __init__(self, offset_sigma=10.0, offset_scale=5.0, **kwargs):
        super().__init__(**kwargs)
        try:
            offset_sigma[0]
        except:
            offset_sigma = (offset_sigma,)

        try:
            offset_scale[0]
        except:
            offset_scale = (offset_scale,)

        self.offset_scale = jnp.array(offset_scale)
        self.offset_sigma = jnp.array(offset_sigma)

    def call(self, inputs, preds, **kwargs):

        return self_supervised_segmentation_loss(
            preds,
            inputs,
            offset_scale=self.offset_scale,
            offset_sigma=self.offset_sigma,
        )



import glob
import json
from functools import partial
from os.path import join

import imageio.v2 as imageio
import numpy as np
import tensorflow as tf
from skimage.measure import regionprops


# A adaptation of tf.image.crop_and_resize to use edge-based indexing
# Edge-based indexing is more accurate for float-value bboxes
# i.e. value (0, 0) refers to the top-left cornor (not center) of the top-left pixel
def _crop_and_resize(masks, boxes, target_shape):
    box_h = boxes[:, 2] - boxes[:, 0]
    box_w = boxes[:, 3] - boxes[:, 1]
    dh = box_h / target_shape[0] / 2
    dw = box_w / target_shape[1] / 2
    boxes = tf.stack([dh, dw, -dh, -dw], axis=-1) + boxes - 0.5
    H, W = masks.shape[1:3]
    segs = tf.image.crop_and_resize(
        masks[..., None],
        boxes / [H - 1, W - 1, H - 1, W - 1],
        tf.range(len(boxes)),
        target_shape,
    )
    return tf.squeeze(segs, axis=-1)


def coco_generator_full(annotation_file, image_path, mask_shape=None):
    from pycocotools.coco import COCO

    coco = COCO(annotation_file=annotation_file)
    for imgid in coco.getImgIds():
        imginfo = coco.imgs[imgid]
        bboxes = []
        locs = []
        masks = []
        # segs = []
        for ann_id in coco.getAnnIds(imgIds=imgid):
            ann = coco.anns[ann_id]
            bbox = ann["bbox"]
            bbox = np.array([bbox[1], bbox[0], bbox[1] + bbox[3], bbox[0] + bbox[2]])
            bboxes.append(bbox)

            mask = coco.annToMask(ann)
            masks.append(mask)

            mi = np.stack(np.where(mask >= 0.5), axis=-1)
            locs.append(mi.mean(axis=0) + 0.5)

        bboxes = np.array(bboxes, dtype="float32")
        locs = np.array(locs, dtype="float32")
        masks = np.array(masks, dtype="float32")

        filepath = glob.glob(join(image_path, "**", imginfo["file_name"]))
        assert len(filepath) == 1
        image = imageio.imread(filepath[0])

        if len(image.shape) == 2:
            image = image[:, :, None]
        image = (image / 255.0).astype("float32")

        img_h, img_w, _ = image.shape
        bboxes = np.clip(
            bboxes / [img_h, img_w, img_h, img_w],
            0.0,
            1.0,
        ) * [img_h, img_w, img_h, img_w]
        bboxes = bboxes.astype("float32")

        locs = np.clip(
            locs / [img_h, img_w],
            0.0,
            1.0,
        ) * [img_h, img_w]
        locs = locs.astype("float32")

        if mask_shape is not None:
            segs = _crop_and_resize(masks, bboxes, mask_shape).numpy()
        else:
            segs = masks

        n_cells = masks.shape[0]
        masks = masks.astype("int32") * np.arange(1, n_cells + 1).reshape(n_cells, 1, 1)
        label = masks.max(axis=0).astype("int32")

        yield {
            "id": imgid,
            "filename": imginfo["file_name"],
            "image": image,
            "masks": segs,
            "centroids": locs,
            "bboxes": bboxes,
            "label": label,
        }


def coco_generator(annotation_file, image_path):
    from pycocotools.coco import COCO

    coco = COCO(annotation_file=annotation_file)
    for imgid in coco.getImgIds():
        imginfo = coco.imgs[imgid]
        bboxes = []
        segs = []
        for ann_id in coco.getAnnIds(imgIds=imgid):
            ann = coco.anns[ann_id]
            bbox = ann["bbox"]
            bbox = [bbox[1], bbox[0], bbox[1] + bbox[3], bbox[0] + bbox[2]]
            bboxes.append(bbox)

            seg = np.reshape(coco.anns[ann_id]["segmentation"], [-1, 2])
            segs.append(seg)

        filepath = glob.glob(join(image_path, "**", imginfo["file_name"]))
        assert len(filepath) == 1

        image = imageio.imread(filepath[0])

        if len(image.shape) == 2:
            image = image[:, :, None]
        image = (image / 255.0).astype("float32")

        yield {
            "id": imgid,
            "filename": imginfo["file_name"],
            "image": image,
            "bboxes": bboxes,
            "polygons": segs,
        }


def dataset_from_coco_annotations(
    annotation_file,
    image_path,
    image_shape=[None, None, 3],
    mask_shape=[48, 48],
):
    if mask_shape is None:
        mask_spec = tf.TensorSpec([None] + image_shape[:2], dtype=tf.float32)
    else:
        mask_spec = tf.TensorSpec([None] + mask_shape, dtype=tf.float32)

    return tf.data.Dataset.from_generator(
        lambda: coco_generator_full(annotation_file, image_path, mask_shape=mask_shape),
        output_signature={
            "id": tf.TensorSpec([], dtype=tf.int64),
            "filename": tf.TensorSpec([], dtype=tf.string),
            "image": tf.TensorSpec(image_shape, dtype=tf.float32),
            "masks": mask_spec,
            "centroids": tf.TensorSpec([None, 2], dtype=tf.float32),
            "bboxes": tf.TensorSpec([None, 4], dtype=tf.float32),
            "label": tf.TensorSpec(image_shape[:2], dtype=tf.int32),
        },
    )


def simple_generator(annotation_file, image_path):
    with open(annotation_file, "r") as f:
        annotations = json.load(f)

    for k, ann in enumerate(annotations):
        image = imageio.imread(join(image_path, ann["image_file"]))
        if len(image.shape) == 2:
            image = image[:, :, None]
        image = (image / 255.0).astype("float32")

        binary_mask = imageio.imread(join(image_path, ann["mask_file"]))

        if len(binary_mask.shape) == 3:
            binary_mask = binary_mask[:, :, 0]
        binary_mask = (binary_mask > 0).astype("float32")

        locations = np.array(ann["locations"]).astype("float32")

        if "img_id" in ann:
            img_id = ann["img_id"]
        else:
            img_id = k

        yield {
            "img_id": img_id,
            "image": image,
            "image_mask": binary_mask,
            "centroids": locations,
        }


def dataset_from_simple_annotations(
    annotation_file, image_path, image_shape=[None, None, 3], **kwargs
):
    return tf.data.Dataset.from_generator(
        lambda: simple_generator(annotation_file, image_path, **kwargs),
        output_signature={
            "img_id": tf.TensorSpec([], dtype=tf.int64),
            "image": tf.TensorSpec(image_shape, dtype=tf.float32),
            "image_mask": tf.TensorSpec(image_shape[:2], dtype=tf.float32),
            "centroids": tf.TensorSpec([None, 2], dtype=tf.float32),
        },
    )


def img_mask_pair_generator(ds_files):
    for k, (img_file, mask_file) in enumerate(ds_files):
        img = imageio.imread(img_file).astype("float32")
        img /= 255
        if len(img.shape) == 2:
            img = img[:, :, None]

        mask = imageio.imread(mask_file)
        if len(mask.shape) == 3:
            mask = mask[:, :, 0]

        bboxes = []
        locs = []
        for prop in regionprops(mask):
            bboxes.append(prop["bbox"])
            locs.append(prop["centroid"])

        yield {
            "img_id": k,
            "image": img.astype("float32"),
            "centroids": np.asarray(locs, dtype="float32") + .5,
            "bboxes": np.asarray(bboxes),
            "label": mask,
        }


def dataset_from_img_mask_pairs(
    imgfiles, maskfiles, image_shape=[None, None, 3], **kwargs
):
    return tf.data.Dataset.from_generator(
        lambda: img_mask_pair_generator(zip(imgfiles, maskfiles), **kwargs),
        output_signature={
            "img_id": tf.TensorSpec([], dtype=tf.int64),
            "image": tf.TensorSpec(image_shape, dtype=tf.float32),
            "centroids": tf.TensorSpec([None, 2], dtype=tf.float32),
            "bboxes": tf.TensorSpec([None, 4], dtype=tf.float32),
            "label": tf.TensorSpec(image_shape[:2], dtype=tf.int32),
        },
    )



import typing as tp
from dataclasses import asdict, field
from functools import partial

import flax.linen as nn
import jax
import jax.numpy as jnp

from ..ops import *
from .auxiliary import *
from .convnext import ConvNeXt
from .detector import Detector
from .lpn import LPN
from .resnet import ResNet
from .segmentor import Segmentor


class Lacss(nn.Module):
    backbone: nn.Module = ConvNeXt()
    lpn: nn.Module = LPN()
    detector: nn.Module = Detector()
    segmentor: nn.Module = Segmentor()

    @classmethod
    def from_config(cls, config):
        config_dict = {}
        if "backbone" in config:
            config_dict["backbone"] = ConvNeXt(**config["backbone"])
        if "lpn" in config:
            config_dict["lpn"] = LPN(**config["lpn"])
        if "detector" in config:
            config_dict["detector"] = Detector(**config["detector"])
        if "segmentor" in config:
            config_dict["segmentor"] = Segmentor(**config["segmentor"])

        return cls(**config_dict)

    def __call__(
        self,
        image: jnp.ndarray,
        gt_locations: jnp.ndarray = None,
        *,
        training: bool = None,
    ) -> dict:
        """
        Args:
            image: [H, W, C]
            gt_locations: [M, 2] if training, otherwise None
        Returns:
            a dict of model outputs
        """
        orig_height, orig_width, ch = image.shape
        if ch == 1:
            image = jnp.repeat(image, 3, axis=-1)
        elif ch == 2:
            image = jnp.concatenate([image, jnp.zeros_like(image[..., :1])], axis=-1)

        assert image.shape[-1] == 3

        # ensure input size is multiple of 32
        height = ((orig_height - 1) // 32 + 1) * 32
        width = ((orig_width - 1) // 32 + 1) * 32
        image = jnp.pad(
            image, [[0, height - orig_height], [0, width - orig_width], [0, 0]]
        )

        # backbone
        encoder_features, features = self.backbone(image, training=training)
        model_output = dict(
            encoder_features=encoder_features,
            decoder_features=features,
        )

        # detection
        scaled_gt_locations = (
            gt_locations / jnp.array([height, width])
            if gt_locations is not None
            else None
        )
        model_output.update(
            self.lpn(
                inputs=features,
                scaled_gt_locations=scaled_gt_locations,
            )
        )

        if gt_locations is not None or not training:

            model_output.update(
                self.detector(
                    scores=model_output["lpn_scores"],
                    regressions=model_output["lpn_regressions"],
                    gt_locations=gt_locations,
                    training=training,
                )
            )

            # segmentation
            locations = model_output[
                "training_locations" if training else "pred_locations"
            ]
            scaled_locs = locations / jnp.array([height, width])
            model_output.update(
                self.segmentor(
                    features=features,
                    locations=scaled_locs,
                )
            )

        return model_output


class LacssWithHelper(nn.Module):
    cfg: dict = field(default_factory=dict)
    aux_edge_cfg: tp.Optional[dict] = None
    aux_fg_cfg: tp.Optional[dict] = None

    def setup(self):
        self._lacss = Lacss.from_config(self.cfg)
        if self.aux_edge_cfg is not None:
            self._aux_edge_module = AuxInstanceEdge(**self.aux_edge_cfg)
        if self.aux_fg_cfg is not None:
            self._aux_fg_module = AuxForeground(**self.aux_fg_cfg)

    def __call__(self, image, gt_locations=None, category=None, *, training=False):

        preds = self._lacss(image, gt_locations, training=training)

        if self.aux_edge_cfg is not None:

            preds.update(self._aux_edge_module(image, category=category))

        if self.aux_fg_cfg is not None:

            preds.update(
                self._aux_fg_module(image, category=category)
            )

        return preds



""" Various functions deals with segmentation pathces

    All functions here takes unbatched input. Use vmap to convert to batched data
"""

from functools import partial
from typing import Dict, Optional, Sequence, Tuple, Union

import jax
import jax.numpy as jnp
import numpy as np

from .boxes import box_iou_similarity
from .image import sub_pixel_samples


def gather_patches(
    features: jnp.ndarray, locations: jnp.ndarray, patch_size: int
) -> tuple:
    """extract feature patches according to a list of locations

    Args:
        features: [H,W,C] standard 2D feature map
        locations: [N, 2] float32, scaled 0..1
        patch_size: int

    Returns:
        patches: [N, patch_size, patch_size, C]
        y0: [N]: y0 coordinates of patches
        x0: [N]: x0 coordinates of patches
    """
    height, width, _ = features.shape

    locations *= jnp.array([height, width])
    # i_locations = (locations + .5).astype(int)
    i_locations = locations.astype(int)
    i_locations_x = jnp.clip(i_locations[:, 1], 0, width - 1)
    i_locations_y = jnp.clip(i_locations[:, 0], 0, height - 1)
    yy, xx = jnp.mgrid[:patch_size, :patch_size] - patch_size // 2
    xx += i_locations_x[:, None, None]
    yy += i_locations_y[:, None, None]
    remainder = locations - jnp.stack([i_locations_y, i_locations_x], axis=-1)

    # padding to avoid out-of-bound
    padding_size = patch_size // 2 + 1
    paddings = [
        [padding_size, padding_size],
        [padding_size, padding_size],
        [0, 0],
    ]
    padded_features = jnp.pad(features, paddings)

    patches = padded_features[yy + padding_size, xx + padding_size, :]

    return patches, yy[:, 0, 0], xx[:, 0, 0], remainder


def _get_patch_data(pred):
    if isinstance(pred, dict):
        patches = pred["instance_output"]
        yc = pred["instance_yc"]
        xc = pred["instance_xc"]
    else:
        patches, yc, xc = pred

    if patches.ndim > yc.ndim:
        patches = patches.squeeze(-1)

    return patches, yc, xc


def bboxes_of_patches(
    pred: Union[Sequence, Dict], threshold: float = 0.5
) -> jnp.ndarray:
    """Compute the instance bboxes from model predictions

    Args:
        pred: A model prediction dictionary:
        threshold: for segmentation, default 0.5

    Returns:
        bboxes: [n, 4] bboox for empty patches are filled with -1
    """
    patches, yy, xx = _get_patch_data(pred)

    _, d0, d1 = patches.shape
    row_mask = jnp.any(patches > threshold, axis=1)
    col_mask = jnp.any(patches > threshold, axis=2)

    min_col = row_mask.argmax(axis=1)
    max_col = d1 - row_mask[:, ::-1].argmax(axis=1)

    min_row = col_mask.argmax(axis=1)
    max_row = d0 - col_mask[:, ::-1].argmax(axis=1)

    min_row += yy[:, 0, 0]
    max_row += yy[:, 0, 0]
    min_col += xx[:, 0, 0]
    max_col += xx[:, 0, 0]

    is_valid = row_mask.any(axis=1, keepdims=True)
    bboxes = jnp.stack([min_row, min_col, max_row, max_col], axis=-1)
    bboxes = jnp.where(is_valid, bboxes, -1)

    return bboxes


def indices_of_patches(
    pred: Union[Sequence, Dict],
    input_size: Tuple[int, int] = None,
    threshold: float = 0.5,
) -> tuple:
    """Compute yx coodinates of all segmented instances

    Args:
        pred: A model prediction dictionary
        threshold: float

    Returns:
        [N, 3] array. The first two columns are y-x coordinates. The third
            column is a index value for different instances.
    """
    patches, yy, xx = _get_patch_data(pred)

    indices = patches >= threshold
    yy = yy[indices]
    xx = xx[indices]
    rowids = jnp.argwhere(indices)[:, 0]

    if input_size is not None:
        height, width = input_size
        valid_coords = ((yy >= 0) & (yy < height)) & ((xx >= 0) & (xx < width))
        yy = yy[valid_coords]
        xx = xx[valid_coords]
        rowids = rowids[valid_coords]

    return jnp.stack([yy, xx, rowids], axis=-1)


def iou_patches_and_labels(
    pred: dict, labels: jnp.ndarray, BLOCK_SIZE: int = 128, threshold: float = 0.5
) -> jnp.ndarray:
    """Compute iou between prediction and ground truth label.

    Args:
        pred: A model prediction dictionary
        labels: image label. bg_label = 0
        threshold: for segmentation. default is 0.5

    Returns:
        [n, m] array of iou values.
    """
    patches, yc, xc = _get_patch_data(pred)
    patches = patches >= threshold
    pred_areas = jnp.count_nonzero(patches, axis=(-1, -2))
    labels = labels.astype(int)

    pads = yc.shape[-1] // 2
    padded_labels = jnp.pad(labels, pads, constant_values=0)
    gt_patches = padded_labels[yc + pads, xc + pads]

    max_indices = ((labels.max() - 1) // BLOCK_SIZE + 1) * BLOCK_SIZE
    all_gt_areas = jnp.count_nonzero(
        labels[:, :, None] == jnp.arange(max_indices) + 1, axis=(0, 1)
    )

    ious = []
    # FIXME change to scan
    for k in range(1, labels.max() + 1, BLOCK_SIZE):
        gt_p = gt_patches == jnp.arange(k, k + BLOCK_SIZE).reshape(
            -1, 1, 1, 1
        )  # [B, N, s, s]
        gt_areas = all_gt_areas[k - 1 : k + BLOCK_SIZE - 1]
        intersect = jnp.count_nonzero(gt_p & patches, axis=(-1, -2))  # [B, N]
        ious.append(intersect / (pred_areas + gt_areas[:, None] - intersect + 1.0e-8))

    ious = jnp.concatenate(ious, axis=0).transpose()  # [N, B * b]
    ious = ious[:, : labels.max()]  # this breaks JIT

    return ious


def patches_to_segmentations(
    pred: dict, input_size: Tuple[int, int], threshold: float = 0.5
) -> jnp.ndarray:
    """Expand the predicted patches to the full image size.
    The default model segmentation output shows only a small patch around each instance. This
    function expand each patch to the size of the orginal image.

    Args:
        pred: A model prediction dictionary
        input_size: shape of the input image. Tuple of H, W
        threshold: for segmentation. Default is 0.5.

    Returns:
        segmentations: [n, height, width] n full0-size segmenatation masks.

    """
    patches, yc, xc = _get_patch_data(pred)
    n_patches, patch_size, _ = yc.shape

    page_nums = jnp.arange(n_patches)
    segms = jnp.zeros((n_patches,) + input_size)
    segms = segms.at[page_nums[:, None, None], yc, xc].set(patches)

    return (segms >= threshold).astype(int)


def patches_to_label(
    pred: dict,
    input_size: Tuple[int, int],
    mask: Optional[jnp.ndarray] = None,
    score_threshold: float = 0.5,
    threshold: float = 0.5,
    min_cell_area: int = 0,
) -> jnp.ndarray:
    """convert patch output to the image label

    Args:
        pred: A model prediction dictionary
        input_size: shape of the input image. Tuple of H, W
        mask: boolean indicators masking out unwanted instances. Default is None (all cells)
        score_threshold: otional, min_score to be included. Default is .5.
        threshold: segmentation threshold.  Default .5
        min_cell_area: optional minimal cell area to be plotted, default 0.

    Returns:
        label: [height, width]
    """
    label = jnp.zeros(input_size)
    pr = pred["instance_output"] > threshold
    n_patches, patch_size, _ = pr.shape

    if mask is None:
        mask = jnp.ones([n_patches], dtype=bool)
    mask &= pred["instance_mask"].squeeze(axis=(1, 2))
    if score_threshold > 0 and not "training_locations" in pred:
        mask &= pred["pred_scores"] >= score_threshold
    mask &= jnp.count_nonzero(pr, axis=(1, 2)) > min_cell_area

    pr = (pr > threshold).astype(int) * jnp.arange(1, pr.shape[0] + 1)[:, None, None]
    pr = jnp.where(mask[:, None, None], pr, 0)
    pr = jnp.where(pr == 0, 0, pr.max() + 1 - pr)

    yc = pred["instance_yc"]
    xc = pred["instance_xc"]

    label = label.at[yc, xc].max(pr)

    return label


def ious_of_patches_from_same_image(pred: dict, threshold: float = 0.5) -> jnp.ndarray:
    """Compute IOUs among instances from the same image. Most likely used for nms.

    Args:
        pred: A model prediction dictionary
        threshold: Segmentation threshold. Default is 0.5.

    Returns:
        ious: IOUs as an upper triagle matrix.
    """

    bboxes = bboxes_of_patches(pred)
    box_ious = box_iou_similarity(bboxes, bboxes)
    box_ious = np.triu(box_ious, 1)
    # n_detections = jnp.count_nonzero(pred['instance_mask'])
    # box_ious = box_ious[:n_detections, :n_detections]

    cy, cx = np.where(box_ious > 0)
    pad_size = pred["instance_output"].shape[-1]

    patches = np.asarray(pred["instance_output"]) >= threshold
    patches_y = patches[cy]
    patches_x = patches[cx]
    plt_to = np.zeros([len(cy), pad_size * 3, pad_size * 3], dtype=bool)
    dy = (
        np.asarray(pred["instance_yc"])[cy, 0, 0]
        - np.asarray(pred["instance_yc"])[cx, 0, 0]
        + pad_size
    )
    dx = (
        np.asarray(pred["instance_xc"])[cy, 0, 0]
        - np.asarray(pred["instance_xc"])[cx, 0, 0]
        + pad_size
    )

    for k in range(len(cy)):
        plt_to[k, dy[k] : dy[k] + pad_size, dx[k] : dx[k] + pad_size] = patches_y[k]
    plt_to = plt_to[:, pad_size : pad_size * 2, pad_size : pad_size * 2]
    unions = np.count_nonzero(plt_to & patches_x, axis=(1, 2))

    areas_y = np.count_nonzero(patches_y, axis=(1, 2))
    areas_x = np.count_nonzero(patches_x, axis=(1, 2))

    ious = unions / (areas_x + areas_y - unions + 1e-8)

    mask_ious = box_ious
    mask_ious[(cy, cx)] = ious

    return mask_ious


def non_max_suppress_predictions(pred: dict, iou_threshold: float = 0.6) -> jnp.ndarray:
    """Perform nms on the model prediction based on mask IOU

    Args:
        pred: A model prediction dictionary
        ious_threshold: default 0.6

    Returns:
        mask: boolean mask of cells not supressed
    """

    ious = ious_of_patches_from_same_image(pred)
    mask = ious > iou_threshold

    cnt = 0
    while np.count_nonzero(mask) != cnt:
        cnt = jnp.count_nonzero(mask)
        can_suppress_others = ~mask.any(axis=0)
        suppressed = (mask & can_suppress_others[:, None]).any(axis=0)
        mask = mask & ~suppressed[:, None]

    return ~mask.any(axis=0)


_sampling_op = jax.vmap(partial(sub_pixel_samples, edge_indexing=True))


def rescale_patches(
    pred: dict, scale: float
) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]:
    """Rescale/resize instance outputs in a sub-pixel accurate way.
    If the input image was rescaled, this function take care of rescaling the predictions
    to the orginal coodinates.

    Args:
        pred: A model prediction dictionary
        scale: The scaling value. The function does not take a noop shortcut even if scale is 1.

    Returns: A tuple of three arrays
        patches: a 3D array of the rescaled segmentation patches. The array shape should be different from
            the orignal patches in model predition.
        yy: The y-coordinates of the patches in mesh-grid format
        xx: The x-coordinates of the patches in mesh-grid format
    """

    new_y0 = (pred["instance_yc"][:, 0, 0] + 0.5) * scale  # edge indexing in new scale
    new_x0 = (pred["instance_xc"][:, 0, 0] + 0.5) * scale  # edge indexing in new scale
    patch = pred["instance_output"]
    ps = round(patch.shape[-1] * scale)

    yy, xx = jnp.mgrid[:ps, :ps]
    yy = (
        jnp.floor(new_y0[:, None, None]).astype(int) + yy
    )  # center indexing in new scale
    xx = (
        jnp.floor(new_x0[:, None, None]).astype(int) + xx
    )  # center indexing in new scale

    rel_yy = (yy + 0.5) / scale - pred["instance_yc"][
        :, :1, :1
    ]  # edge indexing in old scale
    rel_xx = (xx + 0.5) / scale - pred["instance_xc"][
        :, :1, :1
    ]  # edge indexing in old scale
    new_patch = _sampling_op(
        patch,
        np.stack([rel_yy, rel_xx], axis=-1),
    )

    return new_patch, yy, xx



import dataclasses
import pathlib
import typing as tp
from functools import lru_cache, partial

import cloudpickle
import flax.linen as nn
import jax
import jax.numpy as jnp
from flax.core.frozen_dict import freeze, unfreeze
from flax.training.train_state import TrainState
from optax import GradientTransformation

from ..utils import Inputs, _get_name
from . import strategy
from .data import *
from .loss import LossLog

# so that multiple calls return the same obj
# this avoids JIT when supplying partial func as args
_cached_partial = lru_cache(partial)

LOSSES = tp.Union[tp.Callable, tp.Sequence[tp.Callable]]


def _get_iterator(g):
    try:
        it = iter(g)
    except:
        it = iter(g())

    return it


class Trainer:
    """ FLAX trainer
    The design is to be minimal but help avoiding certain biolerplate code when train with FLAX

    Attributes:
        model: A Flax module
        losses: A list of loss function (or other callabels). 
        optimizer: An optax optimizer
        seed: RNG seed
        strategy: Training strategy. See lacss.train.strategy module. 
        params: Current model parameters. This is a frozen dict. This is read/write.
        initialized: Whether the model has been initialized with an optimizer and initial weights.
    
    Constructor:
        model: A Flax module
        losses: A loss function (callabels) or a list of loss functions, or None. Default is None.
            These should have the call signiture of: 
                loss = loss_fn(inputs, labels, preds)
            The "preds" is the model output. The "inputs" and "labels" are from the train dataset.
            The model trains on the sum of all losses.
        optimizer: An optax optimzier or None
        seed: integer RNG seed. Default is 42.
        strategy: Training strategy. See lacss.train.strategy module. Default is JIT, which trains
            on a single GPU with unbatched input data. Use VMapped strategy for bathced input. Use
            Distributed for multi-GPU training. Use Eager for debugging (no JIT). You can supply 
            your own stategy class.
    """
    def __init__(
        self,
        model: nn.Module,
        losses: tp.Optional[LOSSES] = None,
        optimizer: GradientTransformation = None,
        seed: int = 42,
        strategy: type = strategy.JIT,
    ):
        self.model = model
        self.losses = losses
        self._loss_weights = None

        self.seed = seed if isinstance(seed, jnp.ndarray) else jax.random.PRNGKey(seed)
        self._optimizer = optimizer

        self._strategy = strategy
        self._initialized = False

        # self.reset()

    def reset(self, loss_weights=None):
        """ Reset internal loss value tracking

        Args:
            loss_weights: Optional weights of individual loss functions. If not None, the
                total loss is weighted sum.
        """
        if self.losses is None:
            raise ValueError(f"No loss functions provided")

        losses = self.losses
        try:
            iter(losses)
        except:
            losses = (losses,)

        if loss_weights is not None:
            self._loss_weights = loss_weights
        else:
            loss_weights = self._loss_weights

        if loss_weights is None:
            loss_weights = (1.0,) * len(losses)

        if len(loss_weights) != len(losses):
            raise ValueError(
                f"Loss weights supplied {loss_weights} does not match the number of loss functions ({len(losses)})"
            )

        self.loss_logs = tuple(
            LossLog(loss, w) for loss, w in zip(losses, loss_weights)
        )

    @property
    def initialized(self):
        return self._initialized

    def initialize(self, dataset, tx: GradientTransformation = None):
        """ Initialize the model weights and optimizer states.

        Args:
            dataset: An iterator or generator function to supply the training dataset.
                The dataset should yield (inputs, labels) or inputs. In the latter case
                the labels = None. The inputs is either a tuple or a dict. If the inputs 
                is a dict, the keys are interpreted as the names for keyword args of the 
                model's __call__ function.
            tx: Optional optax optimzier for when the object was constructed without an 
                optimizer
        """
        if tx is None:
            tx = self.optimizer

        if not self._initialized:
            if self.model.scope is None:
                peek = next(_get_iterator(dataset))
                inputs, _, _ = unpack_x_y_sample_weight(peek)

                self.seed, key = jax.random.split(self.seed)
                variables = self._strategy.init_fn(key, self.model, inputs)

            else:
                self.model, variables = self.model.unbind()

            self.state = TrainState.create(
                apply_fn=self.model.apply,
                params=variables["params"],
                tx=tx,
            )

        else:
            raise ValueError("Calling initialize() on already initialized Trainer")

        self._initialized = True

    def __call__(self, inputs, *, strategy=None, **kwargs):
        """ Calling the underlying model

        Args:
            inputs: A tuple or a dict as the inputs for the model.
            strategy: Optionally supply a differnt calling strategy as the default one
                supplied at the constructor.
            **kwargs: Additional keyword args passed on to the model
        
        Returns:
            Model outputs.
        """
        if strategy is None:
            strategy = self._strategy

        predict_fn = strategy.predict

        state = self.state
        if len(kwargs) > 0:
            state = state.replace(
                apply_fn=_cached_partial(self.state.apply_fn, **kwargs)
            )
        preds = predict_fn(state, inputs)

        return preds

    def compute_loss_log(self):
        return {
            _get_name(loss_log.loss_fn): loss_log.compute()
            for loss_log in self.loss_logs
        }

    def train(self, dataset, strategy=None, rng_cols=None, **kwargs):
        """ Create the training iterator

        Args:
            dataset: An iterator or generator function to supply the training dataset.
                See initialize()
            strategy: Optionally override the default strategy.
            rng_cols: Names of any RNG used by the model. Should be a list of strs.
            **kwargs: Additional keyward args passed to the model. E.g. "training=True"
        
        Returns:
            A iterator. Stepping through the iterator will train the model. The iterator
                itself returns a loss log dict, which are mean loss values for each loss
                function.
        
        Usage example:
            trainer.reset()
            train_it = trainer.train(train_dataset, training=True)
            for k in range(train_steps):
                loss_logs = next(train_it)
                if k % 1000 == 0:
                    print(loss_logs)
        """
        if strategy is None:
            strategy = self._strategy

        if not self._initialized:
            raise ValueError("Try to run uninitialized trainer")

        self.reset()

        self.seed, seed = jax.random.split(self.seed)
        for step, data in enumerate(_get_iterator(dataset)):
            inputs, labels, _ = unpack_x_y_sample_weight(data)

            if rng_cols is not None:
                key = jax.random.fold_in(seed, step)
                keys = jax.random.split(key, len(rng_cols))
                rngs = {name: k for name, k in zip(rng_cols, keys)}
            else:
                rngs = None

            train_fn = strategy.train_step

            state = self.state.replace(
                apply_fn=_cached_partial(self.state.apply_fn, **kwargs)
            )
            state, self.loss_logs, preds = train_fn(
                state, self.loss_logs, inputs, labels, rngs
            )
            self.state = state.replace(apply_fn=self.state.apply_fn)

            batch_logs = self.compute_loss_log()

            yield batch_logs

    def save_model(self, path, sub_module: tp.Optional[str] = None):
        """ Save the model in a pickled file. The pickle is a tuple of
            (module, weights).

        Args:
            path: The file path.
            sub_module: Optionally only save a sub_module of the model
                by specifying the name
        """
        module = self.model
        params = self.params

        if sub_module is not None:
            module = module.bind(dict(params=params))
            module = getattr(module, sub_module)
            module, params = module.unbind()
            params = params["params"]

        if isinstance(path, str):
            path = pathlib.Path(path)

        path.parent.mkdir(parents=True, exist_ok=True)
        with open(path, "wb") as f:
            cloudpickle.dump((module, params), f)

    def checkpoint(self, path):
        """ Make a checkpoint of the trainer. This saves the model as well as
        the training states.

        Args:
            path: The file path.
        """
        if isinstance(path, str):
            path = pathlib.Path(path)

        path.parent.mkdir(parents=True, exist_ok=True)

        with open(path, "wb") as f:
            cloudpickle.dump(self, f)

    @staticmethod
    def from_checkpoint(path):
        """ Restore from the checkpoint.

        Args:
            path: The checkpoint file path.

        Returns:
            A new trainer object.
        """
        if isinstance(path, str):
            path = pathlib.Path(path)

        try:
            _bytes = path.read_bytes()
        except BaseException as e:
            raise OSError(f"Could not load the checkpoint. Got exception: {e}")

        trainer = cloudpickle.loads(_bytes)

        if not isinstance(trainer, Trainer):
            raise TypeError("The saved obj is not a Trainer checkpoint")

        return trainer

    def test(self, dataset, metrics, strategy=None):
        """ Create test/validation iterator.

        Args:
            dataset: An iterator or generator function to supply the testing data.
                The iterator should yield a tupple of (inputs, labels). The labels
                should be a dict.
            metrics: A list of Metric objects. They should have two functions:
                m.update(preds, **kwargs):
                    preds is the model output. the remaining kwargs are content of
                    labels.
                m.compute():
                    which should return the accumulated metric value.
            strategy: Optionally override the default strategy.
        
        Returns:
            An iterator. Stepping through it will drive the updating of each metric
                obj. The iterator itself return the list of metrics.
        """
        if strategy is None:
            strategy = self._strategy

        try:
            iter(metrics)
        except TypeError:
            metrics = [metrics]

        # self.initialize(dataset, strategy)

        for data in _get_iterator(dataset):
            inputs, labels, _ = unpack_x_y_sample_weight(data)
            predict_fn = strategy.predict
            preds = predict_fn(self.state, inputs)
            kwargs = dict(
                inputs=inputs,
                preds=preds,
                **labels,
            )
            for m in metrics:
                m.update(**kwargs)

            yield metrics

    def test_and_compute(self, *args, **kwargs):
        """ A convient function to compute all metrics.

        Args:
            same as the test() fucntion
        
        Returns:
            A metric dict. Keys are metric names.
        """
        for metrics in self.test(*args, **kwargs):
            pass
        return {_get_name(m): m.compute() for m in metrics}

    @property
    def params(self):
        return self.state.params

    @params.setter
    def params(self, new_params):
        self.state = self.state.replace(params=new_params)

    @property
    def optimizer(self):
        return self._optimizer

    @optimizer.setter
    def optimizer(self, tx):
        self._optimizer = tx

        if self._initialized:
            self.state = TrainState.create(
                apply_fn=self.model.apply,
                params=self.params,
                tx=tx,
            )



import typing as tp

import flax.linen as nn
import jax
import jax.numpy as jnp

from .common import *
from .unet import UNet


class AuxInstanceEdge(nn.Module):
    conv_spec: tp.Sequence[int] = (24, 64, 64)
    n_groups: int = 1
    # share_weights: bool = False

    @nn.compact
    def __call__(
        self, x: jnp.ndarray, *, category: tp.Optional[jnp.ndarray] = None
    ) -> jnp.ndarray:
        for n in self.conv_spec:
            x = nn.Conv(n, (3, 3), use_bias=False)(x)
            x = nn.GroupNorm(num_groups=None, group_size=1, use_scale=False)(
                x[None, ...]
            )[0]
            x = jax.nn.relu(x)

        x = nn.Conv(self.n_groups, (3, 3))(x)
        c = category.astype(int).squeeze() if category is not None else 0
        x = x[..., c]

        # x = jax.nn.sigmoid(x)

        return dict(
            edge_pred=x,
        )


class AuxForeground(nn.Module):
    conv_spec: tp.Sequence[int] = (24, 64)
    patch_size: tp.Sequence[int] = 1
    n_groups: int = 1
    share_weights: bool = False
    # use_attention: bool = False

    @nn.compact
    def __call__(
        self,
        x: jnp.ndarray,
        *,
        category: tp.Optional[jnp.ndarray] = None,
    ) -> jnp.ndarray:
        assert category is not None or self.n_groups == 1

        net = UNet(self.conv_spec, self.patch_size)
        _, decoder_out = net(x)

        y = decoder_out[str(net.start_level)]

        c = category.astype(int).squeeze() if category is not None else 0

        fg = nn.Conv(self.n_groups, (3, 3))(y)
        fg = fg[..., c]

        if fg.shape != x.shape[:-1]:
            fg = jax.image.resize(fg, x.shape[:-1], "linear")

        output = dict(fg_pred=fg)

        if self.share_weights:
            edge = nn.Conv(self.n_groups, (3, 3))(y)
            edge = edge[..., c]

            if edge.shape != x.shape[:-1]:
                edge = jax.image.resize(edge, x.shape[:-1], "linear")

            output.update(
                dict(
                    edge_pred=edge,
                )
            )

        return output



import dataclasses
import re
import typing as tp

import jax.numpy as jnp
import numpy as np
from flax import struct

from .ops import bboxes_of_patches

InputLike = tp.Union[tp.Any, tp.Tuple[tp.Any, ...], tp.Dict[str, tp.Any], "Inputs"]


def _unique_name(
    names: tp.Set[str],
    name: str,
):

    if name in names:

        match = re.match(r"(.*?)(\d*)$", name)
        assert match is not None

        name = match[1]
        num_part = match[2]

        i = int(num_part) if num_part else 2
        str_template = f"{{name}}{{i:0{len(num_part)}}}"

        while str_template.format(name=name, i=i) in names:
            i += 1

        name = str_template.format(name=name, i=i)

    names.add(name)
    return name


def _unique_names(
    names: tp.Iterable[str],
    *,
    existing_names: tp.Optional[tp.Set[str]] = None,
) -> tp.Iterable[str]:
    if existing_names is None:
        existing_names = set()

    for name in names:
        yield _unique_name(existing_names, name)


def _lower_snake_case(s: str) -> str:
    s = re.sub(r"(?<!^)(?=[A-Z])", "_", s).lower()
    parts = s.split("_")
    output_parts = []

    for i in range(len(parts)):
        if i == 0 or len(parts[i - 1]) > 1:
            output_parts.append(parts[i])
        else:
            output_parts[-1] += parts[i]

    return "_".join(output_parts)


def _get_name(obj) -> str:
    if hasattr(obj, "name") and obj.name:
        return obj.name
    elif hasattr(obj, "__name__") and obj.__name__:
        return _lower_snake_case(obj.__name__)
    elif hasattr(obj, "__class__") and obj.__class__.__name__:
        return _lower_snake_case(obj.__class__.__name__)
    else:
        raise ValueError(f"Could not get name for: {obj}")


class Inputs(struct.PyTreeNode):
    args: tp.Tuple[tp.Any, ...] = ()
    kwargs: tp.Dict[str, tp.Any] = dataclasses.field(default_factory=dict)

    def update(self, *args, **kwargs):
        tmp = self.kwargs.copy()
        tmp.update(kwargs)
        new_inputs = self.replace(args=self.args + args, kwargs=tmp)
        return new_inputs

    @classmethod
    def from_value(cls, value: InputLike) -> "Inputs":
        if isinstance(value, cls):
            return value
        elif isinstance(value, tuple):
            return cls(args=value)
        elif isinstance(value, dict):
            return cls(kwargs=value)
        else:
            return cls(args=(value,))


def _to_str(p):
    return "".join(p.astype(int).reshape(-1).astype(str).tolist())


def format_predictions(pred, mask=None, encode_patch=True, threshold=0.5):
    """Produce more readable data from model predictions
    Args:
        pred: model output without batch dim
        mask: optional mask selecting cells
        threshold: float patch threshold
    Returns: dict(locations, scores, centroids, bboxes, encodings)
    """

    patches = pred["instance_output"] >= threshold
    yc = pred["instance_yc"]
    xc = pred["instance_xc"]

    bboxes = bboxes_of_patches(pred)

    n_pixels = np.count_nonzero(patches, axis=(1, 2))
    centroids = jnp.stack(
        [
            (patches * yc).sum(axis=(1, 2)) / n_pixels,
            (patches * xc).sum(axis=(1, 2)) / n_pixels,
        ],
        axis=-1,
    )  # centroid

    outputs = dict(
        locations=pred["pred_locations"],
        scores=pred["pred_scores"],
        centroids=centroids,
        bboxes=bboxes,
    )

    is_valid = pred["instance_mask"].squeeze(axis=(-1, -2))
    is_valid &= patches.any(axis=(1, 2))  # no empty patches
    if mask is not None:
        is_valid &= mask

    outputs = {k: np.asarray(v)[is_valid] for k, v in outputs.items()}

    if encode_patch:

        encodings = []
        patches = np.asarray(patches)[is_valid]
        y0 = np.asarray(yc)[is_valid, 0, 0]
        x0 = np.asarray(xc)[is_valid, 0, 0]
        boxes = np.asarray(outputs["bboxes"]) - np.stack([y0, x0, y0, x0], axis=-1)

        for box, patch in zip(boxes, patches):
            roi = patch[box[0] : box[2], box[1] : box[3]]
            encodings.append(_to_str(roi))

        outputs["encodings"] = encodings

    return outputs


def show_images(imgs, locs=None, **kwargs):

    import matplotlib.patches
    import matplotlib.pyplot as plt

    fig, axs = plt.subplots(1, len(imgs), figsize=(4 * len(imgs), 5))
    if len(imgs) == 1:
        axs = [axs]

    for k, img in enumerate(imgs):
        axs[k].imshow(img, **kwargs)
        axs[k].axis("off")
        if locs is not None and locs[k] is not None:
            loc = np.round(locs[k]).astype(int)
            for p in loc:
                c = matplotlib.patches.Circle(
                    (p[1], p[0]), fill=False, edgecolor="white"
                )
                axs[k].add_patch(c)
    plt.tight_layout()

def masks2label(inputs):
    import jax
    boxes = inputs['bboxes'].numpy()
    boxes = np.round(boxes).astype(int)
    label = np.zeros(inputs['image'].shape[:2])
    for k, (m, (y1,x1,y2,x2)) in enumerate(zip(inputs["masks"].numpy(), boxes)):
        if y1 > y2:
            y1, y2 = y2, y1
            m=m[::-1,:]
        if x1 > x2:
            x1, x2 = x2, x1
            m = m[:,::-1]
        patch = jax.image.resize(m, [y2-y1, x2-x1], "linear")
        label[y1:y2,x1:x2] = np.maximum(
            (patch >= 0.5).astype(label.dtype) * (k + 1),
            label[y1:y2,x1:x2],
        )
    return label

def convert_old_cp(cp):
    import lacss
    import pickle
    from pathlib import Path
    from flax.core.frozen_dict import freeze, unfreeze

    cp = Path(cp)
    trainer = lacss.train.Trainer.from_checkpoint(cp)

    model = lacss.modules.Lacss(
        backbone = lacss.modules.ConvNeXt(
            depths=(3,3,27,3),
            drop_path_rate=0.4,
            out_channels=256,
        ),
        lpn = lacss.modules.LPN(
            conv_spec=((256,256,256,256),()),
        ),
        detector = lacss.modules.Detector(
            train_max_output=2560,
            test_max_output=2560,
        ),
        segmentor = lacss.modules.Segmentor(
            conv_spec=((256,256,256),(64,)),
            instance_crop_size=96,
        ),
    )
    
    params = trainer.params['_lacss']
    params = freeze(dict(
        backbone=params['_backbone'],
        lpn=params['_lpn'],
        segmentor=params['_segmentor']
    ))

    with open(cp.with_suffix(".pkl"), "wb") as f:
        pickle.dump((model, params), f)


from typing import List, Sequence, Union

import flax.linen as nn
import jax
import jax.numpy as jnp

from .common import *

""" Implements the convnext encoder. Described in https://arxiv.org/abs/2201.03545
Original implementation: https://github.com/facebookresearch/ConvNeXt
"""


class _Block(nn.Module):
    """ConvNeXt Block.
    Args:
        drop_path (float): Stochastic depth rate. Default: 0.0
        layer_scale_init_value (float): Init value for Layer Scale. Default: 1e-6.
    """

    drop_rate: int = 0.0
    layer_scale_init_value: float = 1e-6
    kernel_size: int = 7

    @nn.compact
    def __call__(self, x: jnp.ndarray, *, training=None) -> jnp.ndarray:
        dim = x.shape[-1]
        ks = self.kernel_size
        scale = self.layer_scale_init_value

        shortcut = x

        x = nn.Conv(dim, (ks, ks), feature_group_count=dim)(x)
        x = nn.LayerNorm(epsilon=1e-6)(x)
        x = nn.Dense(dim * 4)(x)
        x = jax.nn.gelu(x)
        x = nn.Dense(dim)(x)

        if scale > 0:
            gamma = self.param(
                "gamma", lambda rng, shape: scale * jnp.ones(shape), (x.shape[-1])
            )
            x = x * gamma

        deterministic = training is None or not training
        x = DropPath(self.drop_rate)(x, deterministic=deterministic)

        x = x + shortcut

        return x


class ConvNeXt(nn.Module):
    """ConvNeXt
    Args:
        patch_size: for stem default 4
        depths (tuple(int)): Number of blocks at each stage. Default: [3, 3, 9, 3]
        dims (int): Feature dimension at each stage. Default: [96, 192, 384, 768]
        drop_path_rate (float): Stochastic depth rate. Default: 0.
        layer_scale_init_value (float): Init value for Layer Scale. Default: 1e-6.
        out_channels (int): FPN output channels. Default: 256
    """

    patch_size: int = 4
    depths: Sequence[int] = (3, 3, 27, 3)
    dims: Sequence[int] = (96, 192, 384, 768)
    drop_path_rate: float = 0.0
    layer_scale_init_value: float = 1e-6
    out_channels: int = 384

    @nn.compact
    def __call__(self, x: jnp.ndarray, *, training: bool = None) -> jnp.ndarray:
        dp_rate = 0
        outputs = []
        for k in range(len(self.depths)):
            if k == 0:
                ps = self.patch_size
                x = nn.Conv(self.dims[k], (ps, ps), strides=(ps, ps))(x)
                x = nn.LayerNorm(epsilon=1e-6)(x)
            else:
                x = nn.LayerNorm(epsilon=1e-6)(x)
                x = nn.Conv(self.dims[k], (2, 2), strides=(2, 2))(x)

            for _ in range(self.depths[k]):
                x = _Block(dp_rate, self.layer_scale_init_value)(x, training=training)
                dp_rate += self.drop_path_rate / (sum(self.depths) - 1)

            outputs.append(x)

        keys = [str(k + 1 if self.patch_size == 2 else k + 2) for k in range(4)]
        encoder_out = dict(zip(keys, outputs))

        if self.out_channels > 0:
            decoder_out = FPN(self.out_channels)(outputs)
            decoder_out = dict(zip(keys, decoder_out))
        else:
            decoder_out = None

        return encoder_out, decoder_out


# utility funcs

model_urls = {
    "convnext_tiny_1k": "https://dl.fbaipublicfiles.com/convnext/convnext_tiny_1k_224_ema.pth",
    "convnext_small_1k": "https://dl.fbaipublicfiles.com/convnext/convnext_small_1k_224_ema.pth",
    "convnext_base_1k": "https://dl.fbaipublicfiles.com/convnext/convnext_base_1k_224_ema.pth",
    "convnext_large_1k": "https://dl.fbaipublicfiles.com/convnext/convnext_large_1k_224_ema.pth",
    "convnext_tiny_22k": "https://dl.fbaipublicfiles.com/convnext/convnext_tiny_22k_224.pth",
    "convnext_small_22k": "https://dl.fbaipublicfiles.com/convnext/convnext_small_22k_224.pth",
    "convnext_base_22k": "https://dl.fbaipublicfiles.com/convnext/convnext_base_22k_224.pth",
    "convnext_large_22k": "https://dl.fbaipublicfiles.com/convnext/convnext_large_22k_224.pth",
    "convnext_xlarge_22k": "https://dl.fbaipublicfiles.com/convnext/convnext_xlarge_22k_224.pth",
}


def load_weight(jax_model, jax_params, url):
    """Load pretrained convnext models
    specs for pretrained models are:
        tiny: dims=(96, 192, 384, 768), depths=(3,3,9,3)
        small: dims=(96, 192, 384, 768), depths=(3,3,27,3)
        base: dims=(128, 256, 512, 1024), depths=(3,3,27,3)
        large: dims=(192, 384, 768, 1536), depths=(3,3,27,3)
        X-large: dims=(256, 512, 1024, 2048), depths=(3,3,27,3)
    """
    import torch
    from flax.core.frozen_dict import freeze, unfreeze

    def t(m, new_value):
        new_value = jnp.array(new_value)
        assert m.shape == new_value.shape
        m = new_value

    checkpoint = torch.hub.load_state_dict_from_url(url=url, map_location="cpu")
    params = checkpoint["model"]
    jax_params = unfreeze(jax_params)

    cnt = 0
    for i in range(4):
        for k in range(jax_model.depths[i]):
            block = jax_params[f"_Block_{cnt}"]
            t(block["gamma"], params[f"stages.{i}.{k}.gamma"])
            t(block["Conv_0"]["bias"], params[f"stages.{i}.{k}.dwconv.bias"])
            t(
                block["Conv_0"]["kernel"],
                params[f"stages.{i}.{k}.dwconv.weight"].permute(2, 3, 1, 0),
            )
            t(block["LayerNorm_0"]["bias"], params[f"stages.{i}.{k}.norm.bias"])
            t(block["LayerNorm_0"]["scale"], params[f"stages.{i}.{k}.norm.weight"])
            t(block["Dense_0"]["bias"], params[f"stages.{i}.{k}.pwconv1.bias"])
            t(
                block["Dense_0"]["kernel"],
                params[f"stages.{i}.{k}.pwconv1.weight"].transpose(0, 1),
            )
            t(block["Dense_1"]["bias"], params[f"stages.{i}.{k}.pwconv2.bias"])
            t(
                block["Dense_1"]["kernel"],
                params[f"stages.{i}.{k}.pwconv2.weight"].transpose(0, 1),
            )

            cnt += 1

    norm = jax_params[f"LayerNorm_0"]
    t(norm["bias"], params["downsample_layers.0.1.bias"])
    t(norm["scale"], params["downsample_layers.0.1.weight"])

    conv = jax_params[f"Conv_0"]
    t(conv["bias"], params["downsample_layers.0.0.bias"])
    t(conv["kernel"], params["downsample_layers.0.0.weight"].permute(2, 3, 1, 0))

    for i in range(1, 4):
        norm = jax_params[f"LayerNorm_{i}"]
        t(norm["bias"], params[f"downsample_layers.{i}.0.bias"])
        t(norm["scale"], params[f"downsample_layers.{i}.0.weight"])
        conv = jax_params[f"Conv_{i}"]
        t(conv["bias"], params[f"downsample_layers.{i}.1.bias"])
        t(conv["kernel"], params[f"downsample_layers.{i}.1.weight"].permute(2, 3, 1, 0))

    return freeze(jax_params)
